==> labpc-0.2.tar.gz: labpc/dio.c <==
/*
 *	$Header: /home/glenn/src/labpc/RCS/dio.c,v 0.2 1995/10/24 16:32:20 glenn Exp glenn $
 *
 *	$Log: dio.c,v $
 *	Revision 0.2  1995/10/24 16:32:20  glenn
 *	Revision 0.2
 *
 */

#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif MODULE

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>

#include <asm/io.h>
#include <asm/segment.h>

#include "dio.h"
#include "labpc.h"

/* The three different digital IO modes described in manual */
#define SIMPLEIOMODE		0
#define PARALLELPORTMODE	1
#define BIDIRECTIONALMODE	2

#define PORTA			0
#define PORTB			1
#define PORTC			2

#define DIO_WRITE		0
#define DIO_READ		1

#define MODEFLAG	0x80
#define MODESELECTA	0x60
#define PORTAIO		0x10
#define PORTCHIGHIO	0x08
#define MODESELECTB	0x04
#define PORTBIO		0x02
#define PORTCLOWIO	0x01

#define GETDIO( minor ) 	\
	(((minor) >= DAQ_DPORTA && (minor) <= DAQ_DPORTC) ?	\
	 	(board->dio) : NULL )

#define SETMODE( dio )		\
	outb_p( (MODEFLAG | (((dio)->modeA & 3) << 5) |			\
		 ((dio)->rwA << 4) | ((dio)->rwC << 3) |		\
		 (((dio)->modeB & 1) << 2) | ((dio)->rwB << 1) |	\
		 ((dio)->rwC)),						\
	        (dio)->mode_port )

int InitDio( Dio_t *dio, int minor, int portA,
	     int portB, int portC, int mode_port )
{
	dio->portA	= portA;
	dio->portB	= portB;
	dio->portC	= portC;
	dio->mode_port	= mode_port;
	dio->minor	= minor;
	dio->modeA	= SIMPLEIOMODE;
	dio->modeB	= SIMPLEIOMODE;
	dio->rwA	= DIO_WRITE;
	dio->rwB	= DIO_WRITE;
	dio->rwC	= DIO_WRITE;
	dio->status	= 0;
	dio->open	= 0;

	SETMODE( dio );

	return 0;
}

static int write_dio( struct inode *inode, struct file *file,
		      const char *buffer, int count )
{
	Dio_t	*dio;
	int	port;
	unsigned long	port_addr;
	int	minor;
	int	i;

	dprintk(( "LabPC+: write_dio()\n" ));
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( minor )) == NULL)
		return -(EINVAL);
	port = minor - dio->minor;

	switch (port) {
	case PORTA:
		port_addr = dio->portA;
		break;
	case PORTB:
		port_addr = dio->portB;
		break;
	case PORTC:
		port_addr = dio->portC;
		break;
	default:
		return -(EINVAL);
	}

	for (i = 0; i < count; i++)
		outb_p( get_fs_byte( &buffer[i] ), port_addr );

	return i;
}

static int read_dio( struct inode *inode, struct file *file,
			char *buffer, int count )
{
	Dio_t	*dio;
	int	port;
	int	val;
	unsigned long	port_addr;
	int	minor;
	int	i;

	dprintk(( "LabPC+: read_dio()\n" ));
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( minor )) == NULL)
		return -(EINVAL);
	port = minor - dio->minor;

	switch (port) {
	case PORTA:
		port_addr = dio->portA;
		break;
	case PORTB:
		port_addr = dio->portB;
		break;
	case PORTC:
		port_addr = dio->portC;
		break;
	default:
		return -(EINVAL);
	}

	for (i = 0; i < count; i++) {
		val = inb_p( port_addr );
		put_fs_byte( val, &buffer[i] );
	}

	return i;
}

static int ioctl_dio( struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg )
{
	Dio_t	*dio;
	int	port;
	int	stat;
	int	minor;

	dprintk(( "LabPC+: ioctl_dio()\n" ));

	if ((stat = labpc_ioctl_check( cmd, arg )) != 0)
		return stat;
		
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( minor )) == NULL)
		return -(EINVAL);
	port = minor - dio->minor;

	switch (cmd) {
	case DIO_SETBIT:
		if (arg > 7)
			return -(EINVAL);
		if (port != PORTC)
			return -(ENOSYS);

		outb_p( (((arg & 7) << 1) | 1), dio->mode_port );
		break;
	case DIO_RESETBIT:
		if (arg > 7)
			return -(EINVAL);	
		if (port != PORTC)
			return -(ENOSYS);

		outb_p( ((arg & 7) << 1), dio->mode_port );
		break;
	case DIO_SETMODE:
		switch (port) {
		case PORTA:
			if (arg != SIMPLEIOMODE &&
			    arg != PARALLELPORTMODE &&
			    arg != BIDIRECTIONALMODE) {
				return -(EINVAL);
			}
			/* Currently only Mode 0 is supported */
			if (arg != SIMPLEIOMODE)
				return -(ENOSYS);
			dio->modeA = arg;
			SETMODE( dio );
			break;
		case PORTB:
			if (arg != SIMPLEIOMODE &&
			    arg != PARALLELPORTMODE &&
			    arg != BIDIRECTIONALMODE) {
				return -(EINVAL);
			}
			if (arg == BIDIRECTIONALMODE)
				return -(ENOSYS);
			/* Currently only Mode 0 is supported */
			if (arg != SIMPLEIOMODE)
				return -(ENOSYS);
			dio->modeB = arg;
			SETMODE( dio );
			break;
		case PORTC:
			return -(ENOSYS);
			break;
		default:
			return -(EINVAL);
		}
		break;
	default:
		return -(EINVAL);
	}

	return 0;
}

static int open_dio( struct inode *inode, struct file *file )
{
	Dio_t	*dio;
	int	minor;

	dprintk(( "LabPC+: open_dio()\n" ));
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( minor )) == NULL)
		return -(EINVAL);

	dio->open++;

	return 0;
}

static void release_dio( struct inode *inode, struct file *file )
{
	Dio_t	*dio;
	int	minor;

	dprintk(( "LabPC+: release_dio()\n" ));
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( minor )) == NULL)
		return;

	dio->open--;
	labpc_close( inode, file );
}

struct file_operations dio_fops = {
	NULL,				/* seek */
	read_dio,
	write_dio,
	NULL,				/* readdir */
	NULL,				/* select_dio */
	ioctl_dio,
	NULL,				/* mmap */
	open_dio,
	release_dio,
};

==> labpc-0.5.tar.gz: labpc/dio.c <==
/*
 *	$Id: dio.c,v 0.5 1996/11/07 14:09:49 glenn Exp glenn $
 */

#include <linux/config.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>

#include <asm/io.h>
#include <asm/segment.h>

#include "dio.h"
#include "labpc.h"

/* The three different digital IO modes described in manual */
#define SIMPLEIOMODE		0
#define PARALLELPORTMODE	1
#define BIDIRECTIONALMODE	2

#define PORTA			0
#define PORTB			1
#define PORTC			2

#define DIO_WRITE		0
#define DIO_READ		1
#define DIO_NONE		2

#define MODEFLAG	0x80
#define MODESELECTA	0x60
#define PORTAIO		0x10
#define PORTCHIGHIO	0x08
#define MODESELECTB	0x04
#define PORTBIO		0x02
#define PORTCLOWIO	0x01

#define SETMODE( dio )		\
	outb_p( (MODEFLAG | (((dio)->modeA & 3) << 5) |			\
		 ((dio)->rwA << 4) | ((dio)->rwC << 3) |		\
		 (((dio)->modeB & 1) << 2) | ((dio)->rwB << 1) |	\
		 ((dio)->rwC)),						\
	        (dio)->mode_port )

/* Static functions don't show up in ksyms - make external for debugging */
#ifndef STATIC
#define STATIC static
#endif

int InitDio( Dio_t *dio, int minor, int portA,
	     int portB, int portC, int mode_port )
{
	dio->portA	= portA;
	dio->portB	= portB;
	dio->portC	= portC;
	dio->mode_port	= mode_port;
	dio->minor	= minor;
	dio->modeA	= SIMPLEIOMODE;
	dio->modeB	= SIMPLEIOMODE;
	dio->rwA	= DIO_WRITE;
	dio->rwB	= DIO_WRITE;
	dio->rwC	= DIO_WRITE;
	dio->status	= 0;
	dio->open	= 0;
	dio->waitA	= NULL;
	dio->waitB	= NULL;

	SETMODE( dio );

	return 0;
}

STATIC int write_dio( struct inode *inode, struct file *file,
		      const char *buffer, int count )
{
	Dio_t	*dio;
	int	port;
	unsigned long	port_addr;
	int	minor;
	int	i;

	dprintk(( "LabPC+: write_dio()\n" ));
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( inode->i_rdev )) == NULL)
		return -(EINVAL);
	port = minor - dio->minor;

	switch (port) {
	case PORTA:
		port_addr = dio->portA;
		if (dio->rwA != DIO_WRITE) {
			dio->rwA = DIO_WRITE;
			SETMODE( dio );
		}
		break;
	case PORTB:
		port_addr = dio->portB;
		if (dio->rwB != DIO_WRITE) {
			dio->rwB = DIO_WRITE;
			SETMODE( dio );
		}
		break;
	case PORTC:
		port_addr = dio->portC;
		if (dio->rwC != DIO_WRITE) {
			dio->rwC = DIO_WRITE;
			SETMODE( dio );
		}
		break;
	default:
		return -(EINVAL);
	}

	for (i = 0; i < count; i++)
		outb_p( get_fs_byte( &buffer[i] ), port_addr );

	return i;
}

STATIC int read_dio( struct inode *inode, struct file *file,
		     char *buffer, int count )
{
	Dio_t	*dio;
	int	port;
	int	val;
	unsigned long	port_addr;
	int	minor;
	int	i;

	dprintk(( "LabPC+: read_dio()\n" ));
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( inode->i_rdev )) == NULL)
		return -(EINVAL);
	port = minor - dio->minor;

	switch (port) {
	case PORTA:
		port_addr = dio->portA;
		if (dio->rwA != DIO_READ) {
			dio->rwA = DIO_READ;
			SETMODE( dio );
		}
		break;
	case PORTB:
		port_addr = dio->portB;
		if (dio->rwB != DIO_READ) {
			dio->rwB = DIO_READ;
			SETMODE( dio );
		}
		break;
	case PORTC:
		port_addr = dio->portC;
		if (dio->rwC != DIO_READ) {
			dio->rwC = DIO_READ;
			SETMODE( dio );
		}
		break;
	default:
		return -(EINVAL);
	}

	for (i = 0; i < count; i++) {
		val = inb_p( port_addr );
		put_fs_byte( val, &buffer[i] );
	}

	return i;
}

STATIC int ioctl_dio( struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg )
{
	Dio_t	*dio;
	int	port;
	int	stat;
	int	minor;

	dprintk(( "LabPC+: ioctl_dio()\n" ));

	if ((stat = labpc_ioctl_check( cmd, arg )) != 0)
		return stat;
		
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( inode->i_rdev )) == NULL)
		return -(EINVAL);
	port = minor - dio->minor;

	switch (cmd) {
	case DIO_SETBIT:
		if (arg > 7)
			return -(EINVAL);
		if (port != PORTC)
			return -(ENOSYS);

		outb_p( (((arg & 7) << 1) | 1), dio->mode_port );
		break;
	case DIO_RESETBIT:
		if (arg > 7)
			return -(EINVAL);	
		if (port != PORTC)
			return -(ENOSYS);

		outb_p( ((arg & 7) << 1), dio->mode_port );
		break;
	case DIO_SETMODE:
		if (arg != SIMPLEIOMODE &&
		    arg != PARALLELPORTMODE &&
		    arg != BIDIRECTIONALMODE) {
			return -(EINVAL);
		}
		switch (port) {
		case PORTA:
			/* Currently only Mode 0 is supported */
			if (arg != SIMPLEIOMODE)
				return -(ENOSYS);
			dio->modeA = arg;
			SETMODE( dio );
			break;
		case PORTB:
			/* Currently only Mode 0 is supported */
			if (arg != SIMPLEIOMODE)
				return -(ENOSYS);
			dio->modeB = arg;
			SETMODE( dio );
			break;
		case PORTC:
			return -(ENOSYS);
			break;
		default:
			return -(EINVAL);
		}
		break;
	default:
		return -(EINVAL);
	}

	return 0;
}

STATIC int select_dio( struct inode *inode, struct file *file,
		       int sel_type, select_table *wait )
{
	Dio_t	*dio;
	int	port,	minor,	status;

	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( inode->i_rdev )) == NULL)
		return -(EINVAL);

	if (sel_type != SEL_IN)
		return 1;

	port = minor - dio->minor;

	status = 0;
	switch (port) {
	case PORTA:
		cli();
		if (inb_p( dio->portC ) & DIO_INTRA) {
			/* Interrupt bits already set - don't wait */
			status = 1;
		} else {
			/* Interrupt bits not set - enable irq and wait */
			setbits( &(dio->board->regs.cmd3), DIOINTEN );
			select_wait( &(dio->waitA), wait );
		}
		sti();
		break;
	case PORTB:
		cli();
		if (inb_p( dio->portC ) & DIO_INTRB) {
			/* Interrupt bits already set - don't wait */
			status = 1;
		} else {
			/* Interrupt bits not set - enable irq and wait */
			setbits( &(dio->board->regs.cmd3), DIOINTEN );
			select_wait( &(dio->waitB), wait );
		}
		sti();
		break;
	default:
		return -(ENOSYS);
	}

	return status;
}

void dio_interrupt( Dio_t *dio, int status )
{
	/* Clear the interrupt or the labpc just keeps generating them */
	clearbits( &(dio->board->regs.cmd3), DIOINTEN );
	if (status & DIO_INTRA) {
		wake_up( &(dio->waitA) );
	}
	if (status & DIO_INTRB) {
		wake_up( &(dio->waitB) );
	}
	return;
}

STATIC int open_dio( struct inode *inode, struct file *file )
{
	Dio_t	*dio;
	int	minor;

	dprintk(( "LabPC+: open_dio()\n" ));
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( inode->i_rdev )) == NULL)
		return -(EINVAL);

	dio->open++;

	return 0;
}

STATIC void release_dio( struct inode *inode, struct file *file )
{
	Dio_t	*dio;
	int	minor;

	dprintk(( "LabPC+: release_dio()\n" ));
	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( inode->i_rdev )) == NULL)
		return;

	dio->open--;
	labpc_close( inode, file );
}

struct file_operations dio_fops = {
	NULL,				/* seek */
	read_dio,
	write_dio,
	NULL,				/* readdir */
	select_dio,
	ioctl_dio,
	NULL,				/* mmap */
	open_dio,
	release_dio,
};
